from astrologic.decorators.if_switcher import switcher as switcher  # noqa: F401
from astrologic.decorators.tail_recursion_optimization import no_recursion as no_recursion  # noqa: F401
from astrologic.decorators.inline import inline as inline  # noqa: F401



import ast
import uuid
import importlib
from typing import List, Dict, Set, Union, Optional

from astrologic.decorators.base import BaseDecorator
from astrologic.function_text import FunctionText


class Inliner(BaseDecorator):
    def change_tree(self, tree, original_function, function_text, *functions, **kwargs):
        all_original_names: Set[str] = self.get_all_names(tree)

        class VisiterCalls(ast.NodeTransformer):
            def visit_Expr(_self, node: ast.Expr) -> Optional[Union[ast.AST, List[ast.AST]]]:
                try:
                    value = node.value
                    _node = node

                    if isinstance(value, ast.Call):
                        node = value
                        if node.func.id in functions:  # type: ignore[attr-defined]
                            function_name = node.func.id  # type: ignore[attr-defined]
                            function_tree = self.get_function_tree(function_name, original_function.__module__)
                            new_block, declarations = self.replace_names(all_original_names, function_tree, node)
                            new_block = new_block.body[0].body
                            return declarations + new_block
                        else:
                            return _node

                except Exception:
                    return node

        return VisiterCalls().visit(tree)

    def get_declaration_block(self, function_tree, call, cache):
        result = []
        args = call.args
        for arg in args:
            old_name = arg.id
            if old_name in cache:
                new_node = ast.Assign(
                    targets=[ast.Name(id=cache[old_name], ctx=ast.Store())],
                    value=arg,
                )
                result.append(new_node)
        tr = function_tree
        for arg, new_arg in zip(args, tr.body[0].args.args):
            old_name = arg.id
            if old_name not in cache:
                new_node = ast.Assign(
                    targets=[ast.Name(id=new_arg.arg, ctx=ast.Store())],
                    value=ast.Name(id=old_name, ctx=ast.Load()),
                )
                result.append(new_node)
        return result

    def replace_names(self, all_original_names: Set[str], tree, call_node):
        cached_names = {}
        allowed = {x.id for x in call_node.args}

        class Visiter(ast.NodeTransformer):
            def visit_Name(_self, node):
                try:
                    name = node.id
                    if name in allowed:
                        new_name = self.new_name(all_original_names, name, cached_names)
                        node.id = new_name
                    return node
                except Exception:
                    return node

        tree = ast.fix_missing_locations(Visiter().visit(tree))
        declarations = self.get_declaration_block(tree, call_node, cached_names)
        return tree, declarations

    def new_name(self, all_original_names: Set[str], name: str, cache: Dict[str, str]):
        if name in cache:
            return cache[name]

        while True:
            new_name = 'generated_' + uuid.uuid4().hex
            if new_name not in all_original_names:
                cache[name] = new_name
                return new_name

    def get_function_tree(self, function_name, module_name) -> Optional[Union[ast.AST, List[ast.AST]]]:
        try:
            module = importlib.import_module(module_name)
        except ModuleNotFoundError:
            return []

        function = getattr(module, function_name)
        text = FunctionText(function)
        tree = self.get_source_tree(text)
        return tree


inline = Inliner()



import ast
import inspect
import functools
from typing import Callable, Dict, Any

import astunparse

from astrologic.decorators.base import BaseDecorator
from astrologic.function_text import FunctionText


class TailRecursionOptimization(BaseDecorator):
    @staticmethod
    def add_prefix_and_postfix(function_text, original_function):
        prefix = """def superfunction(*args, **kwargs):
    is_recursion = False
"""
        postfix = f"""    while True:
        result = {original_function.__name__}(*args, **kwargs)
        if not is_recursion:
            return result
        is_recursion = False
        args, kwargs = result""".replace('\t', function_text.indent_symbols)
        function_text.wrapp_clean_text(prefix=prefix, postfix=postfix, indent=1)

    def get_new_return_node(self, node):
        args = node.value.args
        kwargs = node.value.keywords
        dict_keys = [ast.Constant(value=keyword.arg) for keyword in kwargs]
        dict_values = [keyword.value for keyword in kwargs]
        new_node = ast.Return(
            value=ast.Tuple(
                elts=[
                    ast.Tuple(elts=args, ctx=ast.Load()),
                    ast.Dict(
                        keys=dict_keys,
                        values=dict_values,
                    ),
                ], ctx=ast.Load(),
            ), ctx=ast.Load(),
        )
        return new_node

    def is_recursion(self, function_name: str, node: ast.AST) -> bool:
        try:
            return node.value.func.id == function_name

        except Exception:
            return False

    def convert_tree_to_function(self, tree: ast.AST, original_function: Callable[..., Any], namespace: Dict[str, Any], debug_mode_on: bool) -> Callable[..., Any]:
        if debug_mode_on:
            print('---------------------')
            print(f'Function {original_function.__name__} from module {original_function.__module__} changed. Its code is roughly equivalent to the following:')
            print(astunparse.unparse(tree))
            print('---------------------')

        tree = ast.fix_missing_locations(tree)
        code = compile(tree, filename=inspect.getfile(original_function), mode='exec')
        exec(code, namespace)
        result: Callable[..., Any] = namespace['superfunction']
        result = functools.wraps(original_function)(result)
        return result

    def change_tree(self, tree: ast.AST, original_function: Callable[..., Any], function_text: FunctionText, *args: Any, **kwargs: Any):
        self.add_prefix_and_postfix(function_text, original_function)
        tree = ast.fix_missing_locations(self.get_source_tree(function_text))

        class RewriteName(ast.NodeTransformer):
            def visit_Return(_self, node):
                if self.is_recursion(original_function.__name__, node):
                    _nonlocal = ast.Nonlocal(names=['is_recursion'])
                    flag = ast.Assign(targets=[ast.Name(id='is_recursion', ctx=ast.Store())], value=ast.Constant(value=True))
                    new_return_node = self.get_new_return_node(node)
                    return [_nonlocal, flag, new_return_node]
                return node

        return RewriteName().visit(tree)


no_recursion = TailRecursionOptimization()



import ast
import inspect
import functools
import importlib
from typing import Dict, Set, Callable, Union, Optional, Any

import astunparse

from astrologic.function_text import FunctionText


class BaseDecorator:
    def __call__(self, *args: Callable[..., Any], **kwargs: Any) -> Union[Callable[..., Any], Callable[[Callable[..., Any]], Callable[..., Any]]]:
        def decorator(function: Callable[..., Any]) -> Callable[..., Any]:
            text = FunctionText(function)
            tree = self.get_source_tree(text)
            new_args = () if len(args) == 1 and callable(args[0]) else args
            new_tree = self.change_tree(tree, function, text, *new_args, **kwargs)
            new_tree = ast.fix_missing_locations(new_tree)
            namespace = self.get_new_namespace(function)
            new_function = self.convert_tree_to_function(new_tree, function, namespace, kwargs.get('debug_mode_on'))
            new_function = self.edit_result(function, new_function, namespace)
            return new_function

        if not len(args):
            return decorator
        elif len(args) == 1 and callable(args[0]):
            return decorator(args[0])
        return decorator

    def change_tree(self, tree: ast.AST, function_text: FunctionText, **kwargs: Any) -> ast.AST:
        raise NotImplementedError

    def edit_result(self, original_function: Callable[..., Any], new_function: Callable[..., Any], namespace: Dict[str, Any]) -> Callable[..., Any]:
        return new_function

    def convert_tree_to_function(self, tree: ast.AST, original_function: Callable[..., Any], namespace: Dict[str, Any], debug_mode_on: bool):
        if debug_mode_on:
            print('---------------------')
            print(f'Function {original_function.__name__} from module {original_function.__module__} changed. Its code is roughly equivalent to the following:')
            print(astunparse.unparse(tree))
            print('---------------------')

        code = compile(tree, filename=inspect.getfile(original_function), mode='exec')
        exec(code, namespace)
        result = namespace[original_function.__name__]
        result = functools.wraps(original_function)(result)
        return result

    def get_source_tree(self, text: FunctionText) -> ast.AST:
        tree = ast.parse(text.clean_text_with_pre_breaks)
        return tree

    def get_globals_by_function(self, function: Callable[..., Any]) -> Dict[str, Any]:
        module_name = function.__module__
        try:
            module = importlib.import_module(module_name)
        except ModuleNotFoundError:
            return {}
        
        result = {}
        for object_name in dir(module):
            _object = getattr(module, object_name)
            result[object_name] = _object
        return result

    def get_new_namespace(self, function: Callable[..., Any]) -> Dict[str, Any]:
        return self.get_globals_by_function(function)

    def get_all_names(self, tree: ast.AST) -> Set[str]:
        all_original_names = set()

        class Visiter(ast.NodeVisitor):
            def visit(_self, node: ast.AST) -> Optional[ast.AST]:
                try:
                    if isinstance(node, ast.Name):
                        all_original_names.add(node.id)
                except Exception:
                    pass

        Visiter().visit(tree)

        return all_original_names



import inspect
import textwrap


class FunctionText:
    def __init__(self, function):
        self.function_object = function
        self.source = inspect.getsource(function)
        self.begin_line_number = self.get_begin_line_number(function)
        self.clean_text, self.decorators_num = self.get_clean_text_and_number_of_decorators(self.source)
        self.clean_text_with_pre_breaks = self.get_clean_text_with_pre_breaks(self.clean_text, self.begin_line_number, self.decorators_num)
        self.indent_symbols = self.get_indent_symbols(self.clean_text)

    @staticmethod
    def get_clean_text_with_pre_breaks(clean_text, begin_line_number, decorators_num):
        lines = ['\n' for x in range(begin_line_number - 1 + decorators_num)]
        lines.append(clean_text)
        result = ''.join(lines)
        return result

    @staticmethod
    def get_begin_line_number(function):
        return inspect.getsourcelines(function)[1]

    @staticmethod
    def get_clean_text_and_number_of_decorators(text):
        text = textwrap.dedent(text)
        decorators_num = 0
        for string in text.split('\n'):
            if string.startswith('@'):
                decorators_num += 1
            else:
                break
        clean_text = '\n'.join(text.split('\n')[decorators_num:])
        return clean_text, decorators_num

    @staticmethod
    def get_indent_symbols(text):
        lines = text.split('\n')
        for line in lines:
            if not line.startswith('def '):
                indent_len = len(line) - len(line.lstrip(' \t'))
                return line[:indent_len]
        return ''

    def wrapp_clean_text(self, prefix='', postfix='', indent=0):
        if indent:
            indent_text = textwrap.indent(self.clean_text, indent * self.indent_symbols)
        else:
            indent_text = self.clean_text
        self.clean_text = f'{prefix}\n{indent_text}\n{postfix}'
        n_count = prefix.count('\n')
        self.begin_line_number -= n_count
        if self.begin_line_number < 0:
            self.begin_line_number = 0
        self.clean_text_with_pre_breaks = self.get_clean_text_with_pre_breaks(self.clean_text, self.begin_line_number, self.decorators_num)



import ast
from typing import List, Callable, Union, Optional, Any

from astrologic.decorators.base import BaseDecorator
from astrologic.function_text import FunctionText


class IfSwitcher(BaseDecorator):
    def change_tree(self, tree: ast.AST, original_function: Callable[..., Any], function_text: FunctionText, *args: Any, **kwargs: Any) -> Optional[Union[ast.AST, List[ast.AST]]]:
        class RewriteName(ast.NodeTransformer):
            def visit_If(self, node: ast.If) -> Optional[Union[ast.AST, List[ast.AST]]]:
                try:
                    variable_name = node.test.id
                    if variable_name in kwargs:
                        if not kwargs[variable_name]:
                            return None
                        else:
                            return node.body
                    return node

                except Exception:
                    return node

        return RewriteName().visit(tree)  # type: ignore[no-any-return]


switcher = IfSwitcher()
